@@ -11,18 +11,25 @@ from collections.abc import (
     Iterator,
     Mapping,
     MutableMapping,
+    Set,
 )
 from contextlib import ContextDecorator, contextmanager, suppress
 from dataclasses import dataclass, field
 from enum import Enum, auto
-from functools import partialmethod, singledispatchmethod, wraps
+from functools import (
+    partialmethod,
+    singledispatchmethod,
+    update_wrapper,
+    wraps,
+)
 from inspect import Signature, isclass
-from threading import RLock
-from types import MappingProxyType, UnionType
+from types import UnionType
 from typing import (
     Any,
+    ClassVar,
     ContextManager,
     NamedTuple,
+    NoReturn,
     Protocol,
     TypeVar,
     cast,
@@ -31,18 +38,23 @@ from typing import (
 
 from injection.common.event import Event, EventChannel, EventListener
 from injection.common.lazy import Lazy, LazyMapping
-from injection.common.tools import find_types, format_type, get_origins
+from injection.common.tools.threading import (
+    frozen_collection,
+    synchronized,
+    thread_lock,
+)
+from injection.common.tools.type import find_types, format_type, get_origins
 from injection.exceptions import (
+    InjectionError,
     ModuleError,
     ModuleLockError,
     ModuleNotUsedError,
     NoInjectable,
 )
 
-__all__ = ("Injectable", "Module", "ModulePriorities")
+__all__ = ("Injectable", "Module", "ModulePriority")
 
 _logger = logging.getLogger(__name__)
-_thread_lock = RLock()
 
 _T = TypeVar("_T")
 Types = Iterable[type] | UnionType
@@ -115,7 +127,7 @@ class ModuleRemoved(ModuleEvent):
 @dataclass(frozen=True, slots=True)
 class ModulePriorityUpdated(ModuleEvent):
     module_updated: Module
-    priority: ModulePriorities
+    priority: ModulePriority
 
     def __str__(self) -> str:
         return (
@@ -133,13 +145,15 @@ Injectables
 class Injectable(Protocol[_T]):
     __slots__ = ()
 
-    def __init__(self, __factory: Callable[[], _T] = ..., /): ...
+    def __init__(self, __factory: Callable[[], _T] = None, /):
+        pass
 
     @property
     def is_locked(self) -> bool:
         return False
 
-    def unlock(self): ...
+    def unlock(self):
+        pass
 
     @abstractmethod
     def get_instance(self) -> _T:
@@ -161,7 +175,7 @@ class NewInjectable(BaseInjectable[_T]):
 class SingletonInjectable(BaseInjectable[_T]):
     __slots__ = ("__dict__",)
 
-    __INSTANCE_KEY = "$instance"
+    __INSTANCE_KEY: ClassVar[str] = "$instance"
 
     @property
     def cache(self) -> MutableMapping[str, Any]:
@@ -178,20 +192,58 @@ class SingletonInjectable(BaseInjectable[_T]):
         with suppress(KeyError):
             return self.cache[self.__INSTANCE_KEY]
 
-        with _thread_lock:
+        with thread_lock:
             instance = self.factory()
             self.cache[self.__INSTANCE_KEY] = instance
 
         return instance
 
 
+@dataclass(repr=False, frozen=True, slots=True)
+class ShouldBeInjectable(Injectable[_T]):
+    cls: type[_T]
+
+    def __bool__(self) -> bool:
+        return False
+
+    def get_instance(self) -> NoReturn:
+        raise InjectionError(f"`{format_type(self.cls)}` should be an injectable.")
+
+
+"""
+Broker
+"""
+
+
+@runtime_checkable
+class Broker(Protocol):
+    __slots__ = ()
+
+    @abstractmethod
+    def __getitem__(self, cls: type[_T] | UnionType, /) -> Injectable[_T]:
+        raise NotImplementedError
+
+    @abstractmethod
+    def __contains__(self, cls: type | UnionType, /) -> bool:
+        raise NotImplementedError
+
+    @property
+    @abstractmethod
+    def is_locked(self) -> bool:
+        raise NotImplementedError
+
+    @abstractmethod
+    def unlock(self):
+        raise NotImplementedError
+
+
 """
 Container
 """
 
 
 @dataclass(repr=False, frozen=True, slots=True)
-class Container:
+class Container(Broker):
     __data: dict[type, Injectable] = field(default_factory=dict, init=False)
     __channel: EventChannel = field(default_factory=EventChannel, init=False)
 
@@ -209,26 +261,34 @@ class Container:
     def is_locked(self) -> bool:
         return any(injectable.is_locked for injectable in self.__injectables)
 
+    @property
+    def __classes(self) -> frozenset[type]:
+        return frozenset(self.__data.keys())
+
     @property
     def __injectables(self) -> frozenset[Injectable]:
         return frozenset(self.__data.values())
 
-    def update(self, classes: Types, injectable: Injectable, override: bool):
-        values = MappingProxyType(
-            {origin: injectable for origin in get_origins(*classes)}
-        )
+    def update(self, classes: Iterable[type], injectable: Injectable, override: bool):
+        classes = frozenset(get_origins(*classes))
 
-        if values:
-            event = ContainerDependenciesUpdated(self, values, override)
+        with thread_lock:
+            if not injectable:
+                classes -= self.__classes
+                override = True
 
-            with self.notify(event):
-                if not override:
-                    self.__check_if_exists(*values)
+            if classes:
+                event = ContainerDependenciesUpdated(self, classes, override)
 
-                self.__data.update(values)
+                with self.notify(event):
+                    if not override:
+                        self.__check_if_exists(classes)
+
+                    self.__data.update((cls, injectable) for cls in classes)
 
         return self
 
+    @synchronized
     def unlock(self):
         for injectable in self.__injectables:
             injectable.unlock()
@@ -240,9 +300,11 @@ class Container:
     def notify(self, event: Event) -> ContextManager | ContextDecorator:
         return self.__channel.dispatch(event)
 
-    def __check_if_exists(self, *classes: type):
-        for cls in classes:
-            if cls in self.__data:
+    def __check_if_exists(self, classes: Set[type]):
+        intersection = classes & self.__classes
+
+        for cls in intersection:
+            if self.__data[cls]:
                 raise RuntimeError(
                     f"An injectable already exists for the class `{format_type(cls)}`."
                 )
@@ -253,7 +315,7 @@ Module
 """
 
 
-class ModulePriorities(Enum):
+class ModulePriority(Enum):
     HIGH = auto()
     LOW = auto()
 
@@ -263,7 +325,7 @@ class ModulePriorities(Enum):
 
 
 @dataclass(repr=False, eq=False, frozen=True, slots=True)
-class Module(EventListener):
+class Module(EventListener, Broker):
     name: str = field(default=None)
     __channel: EventChannel = field(default_factory=EventChannel, init=False)
     __container: Container = field(default_factory=Container, init=False)
@@ -283,7 +345,7 @@ class Module(EventListener):
         raise NoInjectable(cls)
 
     def __setitem__(self, cls: type | UnionType, injectable: Injectable, /):
-        self.update((cls,), injectable, override=True)
+        self.update((cls,), injectable)
 
     def __contains__(self, cls: type | UnionType, /) -> bool:
         return any(cls in broker for broker in self.__brokers)
@@ -296,8 +358,8 @@ class Module(EventListener):
         return any(broker.is_locked for broker in self.__brokers)
 
     @property
-    def __brokers(self) -> Iterator[Container | Module]:
-        yield from tuple(self.__modules)
+    def __brokers(self) -> Iterator[Broker]:
+        yield from frozen_collection(self.__modules)
         yield self.__container
 
     def injectable(
@@ -306,11 +368,12 @@ class Module(EventListener):
         /,
         *,
         cls: type[Injectable] = NewInjectable,
+        inject: bool = True,
         on: type | Types = None,
         override: bool = False,
     ):
         def decorator(wp):
-            factory = self.inject(wp, return_factory=True)
+            factory = self.inject(wp, return_factory=True) if inject else wp
             injectable = cls(factory)
             classes = find_types(wp, on)
             self.update(classes, injectable, override)
@@ -320,6 +383,13 @@ class Module(EventListener):
 
     singleton = partialmethod(injectable, cls=SingletonInjectable)
 
+    def should_be_injectable(self, wrapped: type = None, /):
+        def decorator(wp):
+            self[wp] = ShouldBeInjectable(wp)
+            return wp
+
+        return decorator(wrapped) if wrapped else decorator
+
     def set_constant(
         self,
         instance: _T,
@@ -330,6 +400,7 @@ class Module(EventListener):
         cls = type(instance)
         self.injectable(
             lambda: instance,
+            inject=False,
             on=(cls, on),
             override=override,
         )
@@ -340,21 +411,15 @@ class Module(EventListener):
         wrapped: Callable[..., Any] = None,
         /,
         *,
-        force: bool = False,
         return_factory: bool = False,
     ):
         def decorator(wp):
             if not return_factory and isclass(wp):
-                wp.__init__ = self.inject(wp.__init__, force=force)
+                wp.__init__ = self.inject(wp.__init__)
                 return wp
 
-            lazy_binder = Lazy[Binder](lambda: self.__new_binder(wp))
-
-            @wraps(wp)
-            def wrapper(*args, **kwargs):
-                arguments = (~lazy_binder).bind(args, kwargs, force)
-                return wp(*arguments.args, **arguments.kwargs)
-
+            wrapper = InjectedFunction(wp).update(self)
+            self.add_listener(wrapper)
             return wrapper
 
         return decorator(wrapped) if wrapped else decorator
@@ -366,7 +431,7 @@ class Module(EventListener):
             if none:
                 return None
 
-            raise exc from exc
+            raise exc
 
         instance = injectable.get_instance()
         return cast(cls, instance)
@@ -374,14 +439,19 @@ class Module(EventListener):
     def get_lazy_instance(self, cls: type[_T]) -> Lazy[_T | None]:
         return Lazy(lambda: self.get_instance(cls))
 
-    def update(self, classes: Types, injectable: Injectable, override: bool = False):
+    def update(
+        self,
+        classes: Iterable[type],
+        injectable: Injectable,
+        override: bool = False,
+    ):
         self.__container.update(classes, injectable, override)
         return self
 
     def use(
         self,
         module: Module,
-        priority: ModulePriorities = ModulePriorities.get_default(),
+        priority: ModulePriority = ModulePriority.get_default(),
     ):
         if module is self:
             raise ModuleError("Module can't be used by itself.")
@@ -412,13 +482,13 @@ class Module(EventListener):
     def use_temporarily(
         self,
         module: Module,
-        priority: ModulePriorities = ModulePriorities.get_default(),
+        priority: ModulePriority = ModulePriority.get_default(),
     ) -> ContextManager | ContextDecorator:
         self.use(module, priority)
         yield
         self.stop_using(module)
 
-    def change_priority(self, module: Module, priority: ModulePriorities):
+    def change_priority(self, module: Module, priority: ModulePriority):
         event = ModulePriorityUpdated(self, module, priority)
 
         with self.notify(event):
@@ -426,6 +496,7 @@ class Module(EventListener):
 
         return self
 
+    @synchronized
     def unlock(self):
         for broker in self.__brokers:
             broker.unlock()
@@ -454,8 +525,8 @@ class Module(EventListener):
         if self.is_locked:
             raise ModuleLockError(f"`{self}` is locked.")
 
-    def __move_module(self, module: Module, priority: ModulePriorities):
-        last = priority == ModulePriorities.LOW
+    def __move_module(self, module: Module, priority: ModulePriority):
+        last = priority == ModulePriority.LOW
 
         try:
             self.__modules.move_to_end(module, last=last)
@@ -464,44 +535,45 @@ class Module(EventListener):
                 f"`{module}` can't be found in the modules used by `{self}`."
             ) from exc
 
-    def __new_binder(self, target: Callable[..., Any]) -> Binder:
-        signature = inspect.signature(target, eval_str=True)
-        binder = Binder(signature).update(self)
-        self.add_listener(binder)
-        return binder
-
 
 """
-Binder
+InjectedFunction
 """
 
 
 @dataclass(repr=False, frozen=True, slots=True)
 class Dependencies:
-    __mapping: MappingProxyType[str, Injectable]
+    mapping: Mapping[str, Injectable]
 
     def __bool__(self) -> bool:
-        return bool(self.__mapping)
+        return bool(self.mapping)
 
     def __iter__(self) -> Iterator[tuple[str, Any]]:
-        for name, injectable in self.__mapping.items():
+        for name, injectable in self.mapping.items():
             yield name, injectable.get_instance()
 
+    @property
+    def are_resolved(self) -> bool:
+        if isinstance(self.mapping, LazyMapping) and not self.mapping.is_set:
+            return False
+
+        return bool(self)
+
     @property
     def arguments(self) -> OrderedDict[str, Any]:
         return OrderedDict(self)
 
     @classmethod
     def from_mapping(cls, mapping: Mapping[str, Injectable]):
-        return cls(MappingProxyType(mapping))
+        return cls(mapping=mapping)
 
     @classmethod
     def empty(cls):
         return cls.from_mapping({})
 
     @classmethod
-    def resolve(cls, signature: Signature, module: Module):
-        dependencies = LazyMapping(cls.__resolver(signature, module))
+    def resolve(cls, signature: Signature, module: Module, owner: type = None):
+        dependencies = LazyMapping(cls.__resolver(signature, module, owner))
         return cls.from_mapping(dependencies)
 
     @classmethod
@@ -509,33 +581,88 @@ class Dependencies:
         cls,
         signature: Signature,
         module: Module,
+        owner: type = None,
     ) -> Iterator[tuple[str, Injectable]]:
-        for name, parameter in signature.parameters.items():
+        for name, annotation in cls.__get_annotations(signature, owner):
             try:
-                injectable = module[parameter.annotation]
+                injectable = module[annotation]
             except KeyError:
                 continue
 
             yield name, injectable
 
+    @staticmethod
+    def __get_annotations(
+        signature: Signature,
+        owner: type = None,
+    ) -> Iterator[tuple[str, type | Any]]:
+        parameters = iter(signature.parameters.items())
+
+        if owner:
+            name, _ = next(parameters)
+            yield name, owner
+
+        for name, parameter in parameters:
+            yield name, parameter.annotation
+
 
 class Arguments(NamedTuple):
     args: Iterable[Any]
     kwargs: Mapping[str, Any]
 
 
-class Binder(EventListener):
-    __slots__ = ("__signature", "__dependencies")
+class InjectedFunction(EventListener):
+    __slots__ = ("__dict__", "__wrapper", "__dependencies", "__owner")
 
-    def __init__(self, signature: Signature):
-        self.__signature = signature
+    def __init__(self, wrapped: Callable[..., Any], /):
+        update_wrapper(self, wrapped)
+        self.__signature__ = Lazy[Signature](
+            lambda: inspect.signature(wrapped, eval_str=True)
+        )
+
+        @wraps(wrapped)
+        def wrapper(*args, **kwargs):
+            args, kwargs = self.bind(args, kwargs)
+            return wrapped(*args, **kwargs)
+
+        self.__wrapper = wrapper
         self.__dependencies = Dependencies.empty()
+        self.__owner = None
+
+    def __repr__(self) -> str:
+        return repr(self.__wrapper)
+
+    def __str__(self) -> str:
+        return str(self.__wrapper)
+
+    def __call__(self, /, *args, **kwargs) -> Any:
+        return self.__wrapper(*args, **kwargs)
+
+    def __get__(self, instance: object | None, owner: type):
+        if instance is None:
+            return self
+
+        return self.__wrapper.__get__(instance, owner)
+
+    def __set_name__(self, owner: type, name: str):
+        if self.__dependencies.are_resolved:
+            raise TypeError(
+                "`__set_name__` is called after dependencies have been resolved."
+            )
+
+        if self.__owner:
+            raise TypeError("Function owner is already defined.")
+
+        self.__owner = owner
+
+    @property
+    def signature(self) -> Signature:
+        return self.__signature__()
 
     def bind(
         self,
         args: Iterable[Any] = (),
         kwargs: Mapping[str, Any] = None,
-        force: bool = False,
     ) -> Arguments:
         if kwargs is None:
             kwargs = {}
@@ -543,22 +670,25 @@ class Binder(EventListener):
         if not self.__dependencies:
             return Arguments(args, kwargs)
 
-        bound = self.__signature.bind_partial(*args, **kwargs)
+        bound = self.signature.bind_partial(*args, **kwargs)
         dependencies = self.__dependencies.arguments
-
-        if force:
-            bound.arguments |= dependencies
-        else:
-            bound.arguments = dependencies | bound.arguments
+        bound.arguments = dependencies | bound.arguments
 
         return Arguments(bound.args, bound.kwargs)
 
     def update(self, module: Module):
-        self.__dependencies = Dependencies.resolve(self.__signature, module)
+        with thread_lock:
+            self.__dependencies = Dependencies.resolve(
+                self.signature,
+                module,
+                self.__owner,
+            )
+
         return self
 
     @singledispatchmethod
-    def on_event(self, event: Event, /): ...
+    def on_event(self, event: Event, /):
+        pass
 
     @on_event.register
     @contextmanager
